@@ -32,8 +32,10 @@ type Result<T> = result::Result<T, Error>;
 #[derive(Debug)]
 pub enum Command {
   Config,
+  Lock,
   Otp,
   Pin,
+  Pws,
   Status,
   Storage,
 }
@@ -43,8 +45,10 @@ impl Command {
   pub fn execute(&self, args: Vec<String>) -> Result<()> {
     match *self {
       Command::Config => config(args),
+      Command::Lock => lock(args),
       Command::Otp => otp(args),
       Command::Pin => pin(args),
+      Command::Pws => pws(args),
       Command::Status => status(args),
       Command::Storage => storage(args),
     }
@@ -58,8 +62,10 @@ impl fmt::Display for Command {
       "{}",
       match *self {
         Command::Config => "config",
+        Command::Lock => "lock",
         Command::Otp => "otp",
         Command::Pin => "pin",
+        Command::Pws => "pws",
         Command::Status => "status",
         Command::Storage => "storage",
       }
@@ -73,8 +79,10 @@ impl str::FromStr for Command {
   fn from_str(s: &str) -> result::Result<Self, Self::Err> {
     match s {
       "config" => Ok(Command::Config),
+      "lock" => Ok(Command::Lock),
       "otp" => Ok(Command::Otp),
       "pin" => Ok(Command::Pin),
+      "pws" => Ok(Command::Pws),
       "status" => Ok(Command::Status),
       "storage" => Ok(Command::Storage),
       _ => Err(()),
@@ -320,6 +328,54 @@ impl str::FromStr for PinCommand {
   }
 }
 
+#[derive(Debug)]
+enum PwsCommand {
+  Clear,
+  Get,
+  Set,
+  Status,
+}
+
+impl PwsCommand {
+  fn execute(&self, args: Vec<String>) -> Result<()> {
+    match *self {
+      PwsCommand::Clear => pws_clear(args),
+      PwsCommand::Get => pws_get(args),
+      PwsCommand::Set => pws_set(args),
+      PwsCommand::Status => pws_status(args),
+    }
+  }
+}
+
+impl fmt::Display for PwsCommand {
+  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+    write!(
+      f,
+      "{}",
+      match *self {
+        PwsCommand::Clear => "clear",
+        PwsCommand::Get => "get",
+        PwsCommand::Set => "set",
+        PwsCommand::Status => "status",
+      }
+    )
+  }
+}
+
+impl str::FromStr for PwsCommand {
+  type Err = ();
+
+  fn from_str(s: &str) -> result::Result<Self, Self::Err> {
+    match s {
+      "clear" => Ok(PwsCommand::Clear),
+      "get" => Ok(PwsCommand::Get),
+      "set" => Ok(PwsCommand::Set),
+      "status" => Ok(PwsCommand::Status),
+      _ => Err(()),
+    }
+  }
+}
+
 fn parse(parser: &argparse::ArgumentParser<'_>, args: Vec<String>) -> Result<()> {
   if let Err(err) = parser.parse(args, &mut io::stdout(), &mut io::stderr()) {
     Err(Error::ArgparseError(err))
@@ -533,6 +589,15 @@ fn config_set(args: Vec<String>) -> Result<()> {
   commands::config_set(numlock, capslock, scrollock, otp_pin)
 }
 
+/// Lock the Nitrokey.
+fn lock(args: Vec<String>) -> Result<()> {
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Locks the connected Nitrokey device");
+  parse(&parser, args)?;
+
+  commands::lock()
+}
+
 /// Execute an OTP subcommand.
 fn otp(args: Vec<String>) -> Result<()> {
   let mut subcommand = OtpCommand::Get;
@@ -747,6 +812,136 @@ fn pin_unblock(args: Vec<String>) -> Result<()> {
   commands::pin_unblock()
 }
 
+/// Execute a PWS subcommand.
+fn pws(args: Vec<String>) -> Result<()> {
+  let mut subcommand = PwsCommand::Get;
+  let mut subargs = vec![];
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Accesses the password safe");
+  let _ = parser.refer(&mut subcommand).required().add_argument(
+    "subcommand",
+    argparse::Store,
+    "The subcommand to execute (clear|get|set|status)",
+  );
+  let _ = parser.refer(&mut subargs).add_argument(
+    "arguments",
+    argparse::List,
+    "The arguments for the subcommand",
+  );
+  parser.stop_on_first_argument(true);
+  parse(&parser, args)?;
+  drop(parser);
+
+  subargs.insert(0, format!("nitrocli pws {}", subcommand));
+  subcommand.execute(subargs)
+}
+
+/// Access a slot of the password safe on the Nitrokey.
+fn pws_get(args: Vec<String>) -> Result<()> {
+  let mut slot: u8 = 0;
+  let mut name = false;
+  let mut login = false;
+  let mut password = false;
+  let mut quiet = false;
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Reads a password safe slot");
+  let _ = parser.refer(&mut slot).required().add_argument(
+    "slot",
+    argparse::Store,
+    "The PWS slot to read",
+  );
+  let _ = parser.refer(&mut name).add_option(
+    &["-n", "--name"],
+    argparse::StoreTrue,
+    "Show the name stored on the slot",
+  );
+  let _ = parser.refer(&mut login).add_option(
+    &["-l", "--login"],
+    argparse::StoreTrue,
+    "Show the login stored on the slot",
+  );
+  let _ = parser.refer(&mut password).add_option(
+    &["-p", "--password"],
+    argparse::StoreTrue,
+    "Show the password stored on the slot",
+  );
+  let _ = parser.refer(&mut quiet).add_option(
+    &["-q", "--quiet"],
+    argparse::StoreTrue,
+    "Print the stored data without description",
+  );
+  parse(&parser, args)?;
+  drop(parser);
+
+  commands::pws_get(slot, name, login, password, quiet)
+}
+
+/// Set a slot of the password safe on the Nitrokey.
+fn pws_set(args: Vec<String>) -> Result<()> {
+  let mut slot: u8 = 0;
+  let mut name = String::new();
+  let mut login = String::new();
+  let mut password = String::new();
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Writes a password safe slot");
+  let _ = parser.refer(&mut slot).required().add_argument(
+    "slot",
+    argparse::Store,
+    "The PWS slot to read",
+  );
+  let _ = parser.refer(&mut name).required().add_argument(
+    "name",
+    argparse::Store,
+    "The name to store on the slot",
+  );
+  let _ = parser.refer(&mut login).required().add_argument(
+    "login",
+    argparse::Store,
+    "The login to store on the slot",
+  );
+  let _ = parser.refer(&mut password).required().add_argument(
+    "password",
+    argparse::Store,
+    "The password to store on the slot",
+  );
+  parse(&parser, args)?;
+  drop(parser);
+
+  commands::pws_set(slot, &name, &login, &password)
+}
+
+/// Clear a PWS slot.
+fn pws_clear(args: Vec<String>) -> Result<()> {
+  let mut slot: u8 = 0;
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Clears a password safe slot");
+  let _ = parser.refer(&mut slot).required().add_argument(
+    "slot",
+    argparse::Store,
+    "The PWS slot to clear",
+  );
+  parse(&parser, args)?;
+  drop(parser);
+
+  commands::pws_clear(slot)
+}
+
+/// Print the status of the PWS slots.
+fn pws_status(args: Vec<String>) -> Result<()> {
+  let mut all = false;
+  let mut parser = argparse::ArgumentParser::new();
+  parser.set_description("Prints the status of the PWS slots");
+  let _ = parser.refer(&mut all).add_option(
+    &["-a", "--all"],
+    argparse::StoreTrue,
+    "Show slots that are not programmed",
+  );
+  parse(&parser, args)?;
+  drop(parser);
+
+  commands::pws_status(all)
+}
+
 /// Parse the command-line arguments and return the selected command and
 /// the remaining arguments for the command.
 fn parse_arguments(args: Vec<String>) -> Result<(Command, Vec<String>)> {
@@ -757,7 +952,7 @@ fn parse_arguments(args: Vec<String>) -> Result<(Command, Vec<String>)> {
   let _ = parser.refer(&mut command).required().add_argument(
     "command",
     argparse::Store,
-    "The command to execute (config|otp|pin|status|storage)",
+    "The command to execute (config|lock|otp|pin|pws|status|storage)",
   );
   let _ = parser.refer(&mut subargs).add_argument(
     "arguments",
