@@ -1,7 +1,7 @@
 // commands.rs
 
 // *************************************************************************
-// * Copyright (C) 2018 Daniel Mueller (deso@posteo.net)                   *
+// * Copyright (C) 2018-2019 Daniel Mueller (deso@posteo.net)              *
 // *                                                                       *
 // * This program is free software: you can redistribute it and/or modify  *
 // * it under the terms of the GNU General Public License as published by  *
@@ -20,6 +20,9 @@
 use std::fmt;
 use std::result;
 use std::time;
+use std::u8;
+
+use libc::sync;
 
 use nitrokey::ConfigureOtp;
 use nitrokey::Device;
@@ -31,7 +34,7 @@ use crate::pinentry;
 use crate::Result;
 
 /// Create an `error::Error` with an error message of the format `msg: err`.
-fn get_error(msg: &str, err: &nitrokey::CommandError) -> Error {
+fn get_error(msg: &str, err: nitrokey::CommandError) -> Error {
   Error::Error(format!("{}: {:?}", msg, err))
 }
 
@@ -49,6 +52,21 @@ fn get_storage_device() -> Result<nitrokey::Storage> {
   })
 }
 
+/// Open the password safe on the given device.
+fn get_password_safe(device: &dyn Device) -> Result<nitrokey::PasswordSafe<'_>> {
+  try_with_passphrase_and_data(
+    pinentry::PinType::User,
+    "Could not access the password safe",
+    (),
+    |_, passphrase| {
+      device
+        .get_password_safe(passphrase)
+        .map_err(|err| ((), err))
+    },
+  )
+  .map_err(|(_, err)| err)
+}
+
 /// Authenticate the given device using the given PIN type and operation.
 ///
 /// If an error occurs, the error message `msg` is used.
@@ -156,7 +174,7 @@ where
           let error = format!("{}: Wrong password", msg);
           return Err((new_data, Error::Error(error)));
         }
-        err => return Err((new_data, get_error(msg, &err))),
+        err => return Err((new_data, get_error(msg, err))),
       },
     };
   }
@@ -180,7 +198,7 @@ where
 fn print_status(model: &'static str, device: &nitrokey::DeviceWrapper) -> Result<()> {
   let serial_number = device
     .get_serial_number()
-    .map_err(|err| get_error("Could not query the serial number", &err))?;
+    .map_err(|err| get_error("Could not query the serial number", err))?;
   println!(
     r#"Status:
   model:             {model}
@@ -218,11 +236,6 @@ pub fn storage_open() -> Result<()> {
   )
 }
 
-#[link(name = "c")]
-extern "C" {
-  fn sync();
-}
-
 /// Close the previously opened encrypted volume.
 pub fn storage_close() -> Result<()> {
   // Flush all filesystem caches to disk. We are mostly interested in
@@ -233,7 +246,7 @@ pub fn storage_close() -> Result<()> {
 
   get_storage_device()?
     .disable_encrypted_volume()
-    .map_err(|err| get_error("Closing encrypted volume failed", &err))
+    .map_err(|err| get_error("Closing encrypted volume failed", err))
 }
 
 /// Pretty print the status of a Nitrokey Storage.
@@ -269,7 +282,7 @@ pub fn storage_status() -> Result<()> {
   let device = get_storage_device()?;
   let status = device
     .get_status()
-    .map_err(|err| get_error("Getting Storage status failed", &err))?;
+    .map_err(|err| get_error("Getting Storage status failed", err))?;
 
   print_storage_status(&status);
   Ok(())
@@ -287,7 +300,7 @@ fn format_option<T: fmt::Display>(option: Option<T>) -> String {
 pub fn config_get() -> Result<()> {
   let config = get_device()?
     .get_config()
-    .map_err(|err| get_error("Could not get configuration", &err))?;
+    .map_err(|err| get_error("Could not get configuration", err))?;
   println!(
     r#"Config:
   numlock binding:          {nl}
@@ -312,7 +325,7 @@ pub fn config_set(
   let device = authenticate_admin(get_device()?)?;
   let config = device
     .get_config()
-    .map_err(|err| get_error("Could not get configuration", &err))?;
+    .map_err(|err| get_error("Could not get configuration", err))?;
   let config = nitrokey::Config {
     numlock: numlock.or(config.numlock),
     capslock: capslock.or(config.capslock),
@@ -321,7 +334,14 @@ pub fn config_set(
   };
   device
     .write_config(config)
-    .map_err(|err| get_error("Could not set configuration", &err))
+    .map_err(|err| get_error("Could not set configuration", err))
+}
+
+/// Lock the Nitrokey device.
+pub fn lock() -> Result<()> {
+  get_device()?
+    .lock()
+    .map_err(|err| get_error("Getting Storage status failed", err))
 }
 
 fn get_otp<T: GenerateOtp>(slot: u8, algorithm: args::OtpAlgorithm, device: &T) -> Result<String> {
@@ -329,7 +349,7 @@ fn get_otp<T: GenerateOtp>(slot: u8, algorithm: args::OtpAlgorithm, device: &T)
     args::OtpAlgorithm::Hotp => device.get_hotp_code(slot),
     args::OtpAlgorithm::Totp => device.get_totp_code(slot),
   }
-  .map_err(|err| get_error("Could not generate OTP", &err))
+  .map_err(|err| get_error("Could not generate OTP", err))
 }
 
 fn get_unix_timestamp() -> Result<u64> {
@@ -352,11 +372,11 @@ pub fn otp_get(slot: u8, algorithm: args::OtpAlgorithm, time: Option<u64>) -> Re
         Some(time) => time,
         None => get_unix_timestamp()?,
       })
-      .map_err(|err| get_error("Could not set time", &err))?;
+      .map_err(|err| get_error("Could not set time", err))?;
   }
   let config = device
     .get_config()
-    .map_err(|err| get_error("Could not get device configuration", &err))?;
+    .map_err(|err| get_error("Could not get device configuration", err))?;
   let otp = if config.user_password {
     let user = authenticate_user(device)?;
     get_otp(slot, algorithm, &user)
@@ -408,7 +428,7 @@ pub fn otp_set(
     args::OtpAlgorithm::Hotp => device.write_hotp_slot(data, counter),
     args::OtpAlgorithm::Totp => device.write_totp_slot(data, time_window),
   }
-  .map_err(|err| get_error("Could not write OTP slot", &err))?;
+  .map_err(|err| get_error("Could not write OTP slot", err))?;
   Ok(())
 }
 
@@ -419,7 +439,7 @@ pub fn otp_clear(slot: u8, algorithm: args::OtpAlgorithm) -> Result<()> {
     args::OtpAlgorithm::Hotp => device.erase_hotp_slot(slot),
     args::OtpAlgorithm::Totp => device.erase_totp_slot(slot),
   }
-  .map_err(|err| get_error("Could not clear OTP slot", &err))?;
+  .map_err(|err| get_error("Could not clear OTP slot", err))?;
   Ok(())
 }
 
@@ -452,7 +472,7 @@ fn print_otp_status(
           continue;
         }
       }
-      Err(err) => return Err(get_error("Could not check OTP slot", &err)),
+      Err(err) => return Err(get_error("Could not check OTP slot", err)),
     };
     println!("{}\t{}\t{}", algorithm, slot - 1, name);
   }
@@ -532,6 +552,95 @@ pub fn pin_unblock() -> Result<()> {
   )
 }
 
+fn print_pws_data(
+  description: &'static str,
+  result: result::Result<String, nitrokey::CommandError>,
+  quiet: bool,
+) -> Result<()> {
+  let value = result.map_err(|err| get_error("Could not access PWS slot", err))?;
+  if quiet {
+    println!("{}", value);
+  } else {
+    println!("{} {}", description, value);
+  }
+  Ok(())
+}
+
+/// Read a PWS slot.
+pub fn pws_get(
+  slot: u8,
+  show_name: bool,
+  show_login: bool,
+  show_password: bool,
+  quiet: bool,
+) -> Result<()> {
+  let device = get_device()?;
+  let pws = get_password_safe(&device)?;
+  let show_all = !show_name && !show_login && !show_password;
+  if show_all || show_name {
+    print_pws_data("name:    ", pws.get_slot_name(slot), quiet)?;
+  }
+  if show_all || show_login {
+    print_pws_data("login:   ", pws.get_slot_login(slot), quiet)?;
+  }
+  if show_all || show_password {
+    print_pws_data("password:", pws.get_slot_password(slot), quiet)?;
+  }
+  Ok(())
+}
+
+/// Write a PWS slot.
+pub fn pws_set(slot: u8, name: &str, login: &str, password: &str) -> Result<()> {
+  let device = get_device()?;
+  let pws = get_password_safe(&device)?;
+  pws
+    .write_slot(slot, name, login, password)
+    .map_err(|err| get_error("Could not write PWS slot", err))
+}
+
+/// Clear a PWS slot.
+pub fn pws_clear(slot: u8) -> Result<()> {
+  let device = get_device()?;
+  let pws = get_password_safe(&device)?;
+  pws
+    .erase_slot(slot)
+    .map_err(|err| get_error("Could not clear PWS slot", err))
+}
+
+fn print_pws_slot(pws: &nitrokey::PasswordSafe<'_>, slot: usize, programmed: bool) -> Result<()> {
+  if slot > u8::MAX as usize {
+    return Err(Error::Error("Invalid PWS slot number".to_string()));
+  }
+  let slot = slot as u8;
+  let name = if programmed {
+    pws
+      .get_slot_name(slot)
+      .map_err(|err| get_error("Could not read PWS slot", err))?
+  } else {
+    "[not programmed]".to_string()
+  };
+  println!("{}\t{}", slot, name);
+  Ok(())
+}
+
+/// Print the status of all PWS slots.
+pub fn pws_status(all: bool) -> Result<()> {
+  let device = get_device()?;
+  let pws = get_password_safe(&device)?;
+  let slots = pws
+    .get_slot_status()
+    .map_err(|err| get_error("Could not read PWS slot status", err))?;
+  println!("slot\tname");
+  for (i, &value) in slots
+    .into_iter()
+    .enumerate()
+    .filter(|(_, &value)| all || value)
+  {
+    print_pws_slot(&pws, i, value)?;
+  }
+  Ok(())
+}
+
 #[cfg(test)]
 mod tests {
   use super::*;
