@@ -17,11 +17,12 @@
 // * along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 // *************************************************************************
 
+use std::fmt;
 use std::process;
+use std::str;
 
 use crate::error::Error;
 
-
 /// PIN type requested from pinentry.
 ///
 /// The available PIN types correspond to the PIN types used by the Nitrokey devices:  user and
@@ -29,12 +30,35 @@ use crate::error::Error;
 #[derive(Clone, Copy, Debug, PartialEq)]
 pub enum PinType {
   /// The admin PIN.
-  #[allow(unused)]
   Admin,
   /// The user PIN.
   User,
 }
 
+impl fmt::Display for PinType {
+  fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+    write!(
+      f,
+      "{}",
+      match *self {
+        PinType::Admin => "admin",
+        PinType::User => "user",
+      }
+    )
+  }
+}
+
+impl str::FromStr for PinType {
+  type Err = ();
+
+  fn from_str(s: &str) -> Result<Self, Self::Err> {
+    match s {
+      "admin" => Ok(PinType::Admin),
+      "user" => Ok(PinType::User),
+      _ => Err(()),
+    }
+  }
+}
 
 impl PinType {
   fn cache_id(self) -> &'static str {
@@ -51,14 +75,42 @@ impl PinType {
     }
   }
 
-  fn description(self) -> &'static str {
+  fn description(self, mode: Mode) -> &'static str {
     match self {
-      PinType::Admin => "Please enter admin PIN",
-      PinType::User => "Please enter user PIN",
+      PinType::Admin => match mode {
+        Mode::Choose => "Please enter a new admin PIN",
+        Mode::Confirm => "Please confirm the new admin PIN",
+        Mode::Query => "Please enter the admin PIN",
+      },
+      PinType::User => match mode {
+        Mode::Choose => "Please enter a new user PIN",
+        Mode::Confirm => "Please confirm the new user PIN",
+        Mode::Query => "Please enter the user PIN",
+      },
     }
   }
 }
 
+/// PIN entry mode for pinentry.
+///
+/// This enum describes the context of the pinentry query, for example prompting for the current
+/// PIN or requesting a new PIN.  The mode may affect the pinentry description and whether a
+/// quality bar is shown.
+#[derive(Clone, Copy, Debug, PartialEq)]
+pub enum Mode {
+  /// Let the user choose a new PIN.
+  Choose,
+  /// Let the user confirm the previously chosen PIN.
+  Confirm,
+  /// Query an existing PIN.
+  Query,
+}
+
+impl Mode {
+  fn show_quality_bar(self) -> bool {
+    self == Mode::Choose
+  }
+}
 
 fn parse_pinentry_passphrase(response: Vec<u8>) -> Result<Vec<u8>, Error> {
   let string = String::from_utf8(response)?;
@@ -77,39 +129,50 @@ fn parse_pinentry_passphrase(response: Vec<u8>) -> Result<Vec<u8>, Error> {
 
   // Check if we are dealing with a special "ERR " line and report that
   // specially.
-  if lines.len() >= 1 && lines[0].starts_with("ERR ") {
+  if !lines.is_empty() && lines[0].starts_with("ERR ") {
     let (_, error) = lines[0].split_at(4);
     return Err(Error::Error(error.to_string()));
   }
   Err(Error::Error("Unexpected response: ".to_string() + &string))
 }
 
-
 /// Inquire a PIN of the given type from the user.
 ///
 /// This function inquires a PIN of the given type from the user or returns the cached passphrase,
-/// if available.  If an error message is set, it is displayed in the passphrase dialog.
-pub fn inquire_passphrase(pin_type: PinType, error_msg: Option<&str>) -> Result<Vec<u8>, Error> {
+/// if available.  If an error message is set, it is displayed in the passphrase dialog.  The
+/// mode describes the context of the pinentry dialog.  It is used to choose an appropriate
+/// description and to decide whether a quality bar is shown in the dialog.
+pub fn inquire_passphrase(
+  pin_type: PinType,
+  mode: Mode,
+  error_msg: Option<&str>,
+) -> Result<Vec<u8>, Error> {
   let cache_id = pin_type.cache_id();
-  let error_msg = error_msg.map(|msg| msg.replace(" ", "+")).unwrap_or(String::from("+"));
-  let prompt = pin_type.prompt();
-  let description = pin_type.description().replace(" ", "+");
+  let error_msg = error_msg
+    .map(|msg| msg.replace(" ", "+"))
+    .unwrap_or_else(|| String::from("+"));
+  let prompt = pin_type.prompt().replace(" ", "+");
+  let description = pin_type.description(mode).replace(" ", "+");
 
-  let args = vec![cache_id, &error_msg, prompt, &description].join(" ");
-  let command = "GET_PASSPHRASE --data ".to_string() + &args;
+  let args = vec![cache_id, &error_msg, &prompt, &description].join(" ");
+  let mut command = "GET_PASSPHRASE --data ".to_string();
+  if mode.show_quality_bar() {
+    command += "--qualitybar ";
+  }
+  command += &args;
   // We could also use the --data parameter here to have a more direct
   // representation of the passphrase but the resulting response was
   // considered more difficult to parse overall. It appears an error
   // reported for the GET_PASSPHRASE command does not actually cause
   // gpg-connect-agent to exit with a non-zero error code, we have to
   // evaluate the output to determine success/failure.
-  let output = process::Command::new("gpg-connect-agent").arg(command)
+  let output = process::Command::new("gpg-connect-agent")
+    .arg(command)
     .arg("/bye")
     .output()?;
   parse_pinentry_passphrase(output.stdout)
 }
 
-
 fn parse_pinentry_response(response: Vec<u8>) -> Result<(), Error> {
   let string = String::from_utf8(response)?;
   let lines: Vec<&str> = string.lines().collect();
@@ -121,18 +184,17 @@ fn parse_pinentry_response(response: Vec<u8>) -> Result<(), Error> {
   Err(Error::Error("Unexpected response: ".to_string() + &string))
 }
 
-
 /// Clear the cached passphrase of the given type.
 pub fn clear_passphrase(pin_type: PinType) -> Result<(), Error> {
   let command = "CLEAR_PASSPHRASE ".to_string() + pin_type.cache_id();
-  let output = process::Command::new("gpg-connect-agent").arg(command)
+  let output = process::Command::new("gpg-connect-agent")
+    .arg(command)
     .arg("/bye")
     .output()?;
 
   parse_pinentry_response(output.stdout)
 }
 
-
 #[cfg(test)]
 mod tests {
   use super::*;
